// https://unix4lyfe.org/benchmarking/
//
// Read the TSC twice in quick succession and take the difference.
//
// Writes to stdout "x y" pairs where x is the time (in tsc units)
// from the start of the run, and y is the time difference (also in tsc units)
// measured in the sample.

#include <cstdint>
#include <iostream>
#include <set>
#include <vector>

static inline uint64_t rdtscp(void) {
  uint32_t lo, hi;
  asm volatile ("rdtscp"
      : /* outputs */ "=a" (lo), "=d" (hi)
      : /* no inputs */
      : /* clobbers */ "%rcx");
  // Ignore IA32_TSC_AUX in ecx.
  // Clobber rcx instead of ecx on amd64 because upper bits are cleared.
  // Must use asm volatile here or the rdtscp gets optimized out.
  return uint64_t(lo) | (uint64_t(hi) << 32);
}

int main(int argc, char** argv) {
  int num_buckets = 2100;
  int samples_per_bucket = 2*6300;
  if (argc != 1 && argc != 3) {
    std::cerr << "usage: " << argv[0] << " num_buckets samples_per_bucket\n";
    return 1;
  }
  if (argc == 3) {
    num_buckets = atoi(argv[1]);
    samples_per_bucket = atoi(argv[2]);
  }

  std::vector<uint64_t> start_times;
  std::vector<std::set<uint32_t>> buckets;
  uint32_t* samples = new uint32_t[samples_per_bucket];

  for (int i = 0; i < num_buckets; ++i) {
    start_times.push_back(rdtscp());
    for (int j = 0; j < samples_per_bucket; ++j) {
      uint64_t t0 = rdtscp();
      uint64_t t1 = rdtscp();
      samples[j] = uint32_t(t1 - t0);
    }
    buckets.push_back({});
    auto& bucket = buckets.back();
    for (int j = 0; j < samples_per_bucket; ++j) {
      bucket.insert(samples[j]);
    }
  }

  for (int i = 0; i < num_buckets; ++i) {
    uint64_t t = start_times[i] - start_times[0];
    for (uint32_t diff : buckets[i]) {
      std::cout << t << " " << diff << "\n";
    }
  }
  delete [] samples;
  return 0;
}
// vim:set ts=2 sw=2 tw=80 et:
